@@ -1,4 +1,5 @@
 use std::env;
+use std::ffi::OsStr;
 #[cfg(target_family = "unix")]
 use std::fs::OpenOptions;
 use std::io::Write;
@@ -8,14 +9,14 @@ use std::path::{Path, PathBuf};
 use std::process::{self, Command};
 use std::str;
 
-use anyhow::{bail, Context, Result};
+use anyhow::{anyhow, bail, Context, Result};
 use fs_err as fs;
 use path_slash::PathBufExt;
 use serde::Deserialize;
 use target_lexicon::{Architecture, Environment, OperatingSystem, Triple};
 
 use crate::linux::ARM_FEATURES_H;
-use crate::macos::LIBICONV_TBD;
+use crate::macos::{LIBCHARSET_TBD, LIBICONV_TBD};
 
 /// Zig linker wrapper
 #[derive(Clone, Debug, clap::Subcommand)]
@@ -83,11 +84,15 @@ impl Zig {
         let is_macos = target.map(|x| x.contains("macos")).unwrap_or_default();
 
         let rustc_ver = rustc_version::version()?;
+        let zig_version = Zig::zig_version()?;
 
         let filter_linker_arg = |arg: &str| {
             if arg == "-lgcc_s" {
                 // Replace libgcc_s with libunwind
                 return Some("-lunwind".to_string());
+            } else if arg.starts_with("--target=") {
+                // We have already passed target via `-target`
+                return None;
             }
             if (is_arm || is_windows_gnu)
                 && arg.ends_with(".rlib")
@@ -102,16 +107,25 @@ impl Zig {
                     // zig doesn't provide gcc_eh alternative
                     // We use libc++ to replace it on windows gnu targets
                     return Some("-lc++".to_string());
+                } else if arg == "-Wl,-Bdynamic"
+                    && (zig_version.major, zig_version.minor) >= (0, 11)
+                {
+                    // https://github.com/ziglang/zig/pull/16058
+                    // zig changes the linker behavior, -Bdynamic won't search *.a for mingw, but this may be fixed in the later version
+                    // here is a workaround to replace the linker switch with -search_paths_first, which will search for *.dll,*lib first,
+                    // then fallback to *.a
+                    return Some("-Wl,-search_paths_first".to_owned());
                 } else if arg == "-lwindows" || arg == "-l:libpthread.a" || arg == "-lgcc" {
                     return None;
                 } else if arg == "-Wl,--disable-auto-image-base"
                     || arg == "-Wl,--dynamicbase"
                     || arg == "-Wl,--large-address-aware"
+                    || (arg.starts_with("-Wl,") && arg.ends_with("/list.def"))
                 {
                     // https://github.com/rust-lang/rust/blob/f0bc76ac41a0a832c9ee621e31aaf1f515d3d6a5/compiler/rustc_target/src/spec/windows_gnu_base.rs#L23
                     // https://github.com/rust-lang/rust/blob/2fb0e8d162a021f8a795fb603f5d8c0017855160/compiler/rustc_target/src/spec/windows_gnu_base.rs#L22
                     // https://github.com/rust-lang/rust/blob/f0bc76ac41a0a832c9ee621e31aaf1f515d3d6a5/compiler/rustc_target/src/spec/i686_pc_windows_gnu.rs#L16
-                    // zig doesn't support --disable-auto-image-base, --dynamicbase, and --large-address-aware
+                    // zig doesn't support --disable-auto-image-base, --dynamicbase and --large-address-aware
                     return None;
                 }
             } else if arg == "-Wl,--no-undefined-version" {
@@ -143,12 +157,32 @@ impl Zig {
                     return None;
                 } else if is_riscv64 {
                     return Some("-march=generic_rv64".to_string());
-                } else if arg.starts_with("-march=armv8-a")
-                    && target
+                } else if arg.starts_with("-march=armv8-a") {
+                    if target
                         .map(|x| x.starts_with("aarch64-macos"))
                         .unwrap_or_default()
-                {
-                    return Some(arg.replace("armv8-a", "apple_a14"));
+                    {
+                        return Some(arg.replace("armv8-a", "apple_m1"));
+                    } else if target
+                        .map(|x| x.starts_with("aarch64-linux"))
+                        .unwrap_or_default()
+                    {
+                        return Some(
+                            arg.replace("armv8-a", "generic+v8a")
+                                .replace("simd", "neon"),
+                        );
+                    }
+                }
+            }
+            if is_macos {
+                if arg.starts_with("-Wl,-exported_symbols_list,") {
+                    // zig doesn't support -exported_symbols_list arg
+                    // https://clang.llvm.org/docs/ClangCommandLineReference.html#cmdoption-clang-exported_symbols_list
+                    return None;
+                }
+                if arg == "-Wl,-dylib" {
+                    // zig doesn't support -dylib
+                    return None;
                 }
             }
             Some(arg.to_string())
@@ -160,6 +194,14 @@ impl Zig {
                 .and_then(|i| args.get(i + 1));
             matches!(undefined, Some(x) if x == "dynamic_lookup")
         };
+        let should_add_libcharset = |args: &[String]| {
+            // See https://github.com/apple-oss-distributions/libiconv/blob/a167071feb7a83a01b27ec8d238590c14eb6faff/xcodeconfig/libiconv.xcconfig
+            if (zig_version.major, zig_version.minor) >= (0, 12) {
+                args.iter().any(|x| x == "-liconv") && !args.iter().any(|x| x == "-lcharset")
+            } else {
+                false
+            }
+        };
 
         let mut new_cmd_args = Vec::with_capacity(cmd_args.len());
         for arg in cmd_args {
@@ -198,6 +240,9 @@ impl Zig {
                 if has_undefined_dynamic_lookup(&link_args) {
                     link_args.push("-Wl,-undefined=dynamic_lookup".to_string());
                 }
+                if should_add_libcharset(&link_args) {
+                    link_args.push("-lcharset".to_string());
+                }
                 if is_windows_msvc {
                     let new_content = link_args.join("\n");
                     let mut out = Vec::with_capacity((1 + new_content.len()) * 2);
@@ -229,8 +274,17 @@ impl Zig {
             new_cmd_args.push("-Wl,-undefined=dynamic_lookup".to_string());
         }
         if is_macos {
-            if let Some(sdkroot) = Self::macos_sdk_root() {
-                let sdkroot = Path::new(&sdkroot);
+            if should_add_libcharset(cmd_args) {
+                new_cmd_args.push("-lcharset".to_string());
+            }
+            let sdkroot = Self::macos_sdk_root();
+            if (zig_version.major, zig_version.minor) >= (0, 12) {
+                // Looks like zig 0.12.0 requires passing `--sysroot`
+                if let Some(ref sdkroot) = sdkroot {
+                    new_cmd_args.push(format!("--sysroot={}", sdkroot.display()));
+                }
+            }
+            if let Some(ref sdkroot) = sdkroot {
                 new_cmd_args.extend_from_slice(&[
                     "-isystem".to_string(),
                     format!("{}", sdkroot.join("usr").join("include").display()),
@@ -348,6 +402,19 @@ impl Zig {
         Ok(PathBuf::from(zig_env.lib_dir))
     }
 
+    fn add_env_if_missing<K, V>(command: &mut Command, name: K, value: V)
+    where
+        K: AsRef<OsStr>,
+        V: AsRef<OsStr>,
+    {
+        let command_env_contains_no_key =
+            |name: &K| !command.get_envs().any(|(key, _)| name.as_ref() == key);
+
+        if command_env_contains_no_key(&name) && env::var_os(&name).is_none() {
+            command.env(name, value);
+        }
+    }
+
     pub(crate) fn apply_command_env(
         manifest_path: Option<&Path>,
         release: bool,
@@ -366,44 +433,36 @@ impl Zig {
         for (parsed_target, raw_target) in rust_targets.iter().zip(&cargo.target) {
             let env_target = parsed_target.replace('-', "_");
             let zig_wrapper = prepare_zig_linker(raw_target)?;
+
             if is_mingw_shell() {
-                let mut add_env = |name, value| {
-                    if env::var_os(&name).is_none() {
-                        cmd.env(name, value);
-                    }
-                };
                 let zig_cc = zig_wrapper.cc.to_slash_lossy();
                 let zig_cxx = zig_wrapper.cxx.to_slash_lossy();
-                add_env(format!("CC_{env_target}"), &*zig_cc);
-                add_env(format!("CXX_{env_target}"), &*zig_cxx);
-                add_env(
-                    format!("CARGO_TARGET_{}_LINKER", env_target.to_uppercase()),
-                    &*zig_cc,
-                );
+                Self::add_env_if_missing(cmd, format!("CC_{env_target}"), &*zig_cc);
+                Self::add_env_if_missing(cmd, format!("CXX_{env_target}"), &*zig_cxx);
+                if !parsed_target.contains("wasm") {
+                    Self::add_env_if_missing(
+                        cmd,
+                        format!("CARGO_TARGET_{}_LINKER", env_target.to_uppercase()),
+                        &*zig_cc,
+                    );
+                }
             } else {
-                let mut add_env = |name, value| {
-                    if env::var_os(&name).is_none() {
-                        cmd.env(name, value);
-                    }
-                };
-                add_env(format!("CC_{env_target}"), &zig_wrapper.cc);
-                add_env(format!("CXX_{env_target}"), &zig_wrapper.cxx);
-                add_env(
-                    format!("CARGO_TARGET_{}_LINKER", env_target.to_uppercase()),
-                    &zig_wrapper.cc,
-                );
+                Self::add_env_if_missing(cmd, format!("CC_{env_target}"), &zig_wrapper.cc);
+                Self::add_env_if_missing(cmd, format!("CXX_{env_target}"), &zig_wrapper.cxx);
+                if !parsed_target.contains("wasm") {
+                    Self::add_env_if_missing(
+                        cmd,
+                        format!("CARGO_TARGET_{}_LINKER", env_target.to_uppercase()),
+                        &zig_wrapper.cc,
+                    );
+                }
             }
 
-            let mut add_env = |name, value| {
-                if env::var_os(&name).is_none() {
-                    cmd.env(name, value);
-                }
-            };
-            add_env(format!("RANLIB_{env_target}"), &zig_wrapper.ranlib);
+            Self::add_env_if_missing(cmd, format!("RANLIB_{env_target}"), &zig_wrapper.ranlib);
             // Only setup AR when explicitly asked to
             // because it need special executable name handling, see src/bin/cargo-zigbuild.rs
             if enable_zig_ar {
-                add_env(format!("AR_{env_target}"), &zig_wrapper.ar);
+                Self::add_env_if_missing(cmd, format!("AR_{env_target}"), &zig_wrapper.ar);
             }
 
             Self::setup_os_deps(manifest_path, release, cargo)?;
@@ -446,43 +505,32 @@ impl Zig {
                 cmd.env("CARGO_TARGET_APPLIES_TO_HOST", "false");
             }
 
-            // bindgen support
-            if let Ok(lib_dir) = Zig::lib_dir() {
-                // Bindgen env vars
-                let bindgen_env = "BINDGEN_EXTRA_CLANG_ARGS";
-                let bindgen_target_env = format!("{}_{}", bindgen_env, env_target);
-                // Closure to add bindgen env vars
-                let add_bindgen_env = |value| {
-                    if env::var_os(&bindgen_env).is_none()
-                        && env::var_os(&bindgen_target_env).is_none()
-                    {
-                        // Only set bindgen env vars if none were set
-                        cmd.env(bindgen_target_env, value);
-                    }
-                };
+            // Pass options used by zig cc down to bindgen, if possible
+            if let Ok(mut options) = Self::collect_zig_cc_options(&zig_wrapper, raw_target) {
+                if raw_target.contains("apple-darwin") {
+                    // everyone seems to miss `#import <TargetConditionals.h>`...
+                    options.push("-DTARGET_OS_IPHONE=0".to_string());
+                }
 
-                let libc = lib_dir.join("libc");
-                if raw_target.contains("linux") {
-                    if raw_target.contains("musl") {
-                        add_bindgen_env(format!("--sysroot={}", libc.join("musl").display()));
-                    } else if raw_target.contains("gnu") {
-                        add_bindgen_env(format!("--sysroot={}", libc.join("glibc").display()));
-                    }
-                } else if raw_target.contains("windows-gnu") {
-                    add_bindgen_env(format!("--sysroot={}", libc.join("mingw").display()));
-                } else if raw_target.contains("apple-darwin") {
-                    if let Some(sdkroot) = Self::macos_sdk_root() {
-                        add_bindgen_env(format!(
-                            "-I{} -F{} -DTARGET_OS_IPHONE=0",
-                            sdkroot.join("usr").join("include").display(),
-                            sdkroot
-                                .join("System")
-                                .join("Library")
-                                .join("Frameworks")
-                                .display()
-                        ));
+                let escaped_options = shlex::join(options.iter().map(|s| &s[..]));
+
+                // Override bindgen variables to append additional options
+                let bindgen_env = "BINDGEN_EXTRA_CLANG_ARGS";
+                let fallback_value = env::var(bindgen_env);
+                for target in [&env_target[..], parsed_target] {
+                    let name = format!("{bindgen_env}_{target}");
+                    if let Ok(mut value) = env::var(&name).or(fallback_value.clone()) {
+                        if shlex::split(&value).is_none() {
+                            // bindgen treats the whole string as a single argument if split fails
+                            value = shlex::quote(&value).into_owned();
+                        }
+                        if !value.is_empty() {
+                            value.push(' ');
+                        }
+                        value.push_str(&escaped_options);
+                        env::set_var(name, value);
                     } else {
-                        add_bindgen_env(format!("--sysroot={}", libc.join("darwin").display()));
+                        env::set_var(name, escaped_options.clone());
                     }
                 }
             }
@@ -490,6 +538,228 @@ impl Zig {
         Ok(())
     }
 
+    /// Collects compiler options used by `zig cc` for given target.
+    /// Used for the case where `zig cc` cannot be used but underlying options should be retained,
+    /// for example, as in bindgen (which requires libclang.so and thus is independent from zig).
+    fn collect_zig_cc_options(zig_wrapper: &ZigWrapper, raw_target: &str) -> Result<Vec<String>> {
+        #[derive(Debug, PartialEq, Eq)]
+        enum Kind {
+            Normal,
+            Framework,
+        }
+
+        #[derive(Debug)]
+        struct PerLanguageOptions {
+            glibc_minor_ver: Option<u32>,
+            include_paths: Vec<(Kind, String)>,
+        }
+
+        fn collect_per_language_options(program: &Path, ext: &str) -> Result<PerLanguageOptions> {
+            // We can't use `-x c` or `-x c++` because pre-0.11 Zig doesn't handle them
+            let empty_file_path = cache_dir().join(format!(".intentionally-empty-file.{ext}"));
+            fs::write(&empty_file_path, "")?;
+
+            let output = Command::new(program)
+                .arg("-E")
+                .arg(&empty_file_path)
+                .arg("-v")
+                .output()?;
+            // Clang always generates UTF-8 regardless of locale, so this is okay.
+            let stderr = String::from_utf8(output.stderr)?;
+            if !output.status.success() {
+                bail!(
+                    "Failed to run `zig cc -v` with status {}: {}",
+                    output.status,
+                    stderr.trim(),
+                );
+            }
+
+            // Collect some macro definitions from cc1 options. We can't directly use
+            // them though, as we can't distinguish options added by zig from options
+            // added by clang driver (e.g. `__GCC_HAVE_DWARF2_CFI_ASM`).
+            let glibc_minor_ver = if let Some(start) = stderr.find("__GLIBC_MINOR__=") {
+                let stderr = &stderr[start + 16..];
+                let end = stderr
+                    .find(|c: char| !c.is_ascii_digit())
+                    .unwrap_or(stderr.len());
+                stderr[..end].parse().ok()
+            } else {
+                None
+            };
+
+            let start = stderr
+                .find("#include <...> search starts here:")
+                .ok_or_else(|| anyhow!("Failed to parse `zig cc -v` output"))?
+                + 34;
+            let end = stderr
+                .find("End of search list.")
+                .ok_or_else(|| anyhow!("Failed to parse `zig cc -v` output"))?;
+
+            let mut include_paths = Vec::new();
+            for mut line in stderr[start..end].lines() {
+                line = line.trim();
+                let mut kind = Kind::Normal;
+                if line.ends_with(" (framework directory)") {
+                    line = line[..line.len() - 22].trim();
+                    kind = Kind::Framework;
+                } else if line.ends_with(" (headermap)") {
+                    bail!("C/C++ search path includes header maps, which are not supported");
+                }
+                if !line.is_empty() {
+                    include_paths.push((kind, line.to_owned()));
+                }
+            }
+
+            Ok(PerLanguageOptions {
+                include_paths,
+                glibc_minor_ver,
+            })
+        }
+
+        let c_opts = collect_per_language_options(&zig_wrapper.cc, "c")?;
+        let cpp_opts = collect_per_language_options(&zig_wrapper.cxx, "cpp")?;
+
+        // Ensure that `c_opts` and `cpp_opts` are almost identical in the way we expect.
+        if c_opts.glibc_minor_ver != cpp_opts.glibc_minor_ver {
+            bail!(
+                "`zig cc` gives a different glibc minor version for C ({:?}) and C++ ({:?})",
+                c_opts.glibc_minor_ver,
+                cpp_opts.glibc_minor_ver,
+            );
+        }
+        let c_paths = c_opts.include_paths;
+        let mut cpp_paths = cpp_opts.include_paths;
+        let (cpp_pre_len, cpp_post_len) = if cpp_paths.starts_with(&c_paths) {
+            (0, cpp_paths.len() - c_paths.len())
+        } else if cpp_paths.ends_with(&c_paths) {
+            (cpp_paths.len() - c_paths.len(), 0)
+        } else {
+            bail!("C++ search path used by `zig cc` is unexpectedly different from C search path!");
+        };
+        if !cpp_paths[..cpp_pre_len]
+            .iter()
+            .chain(cpp_paths[cpp_paths.len() - cpp_post_len..].iter())
+            .all(|(kind, _)| *kind == Kind::Normal)
+        {
+            bail!("C++ search path used by `zig cc` contains additional special search paths!");
+        }
+
+        // <digression>
+        //
+        // So, why we do need all of these?
+        //
+        // Bindgen wouldn't look at our `zig cc` (which doesn't contain `libclang.so` anyway),
+        // but it does collect include paths from the local clang and feed them to `libclang.so`.
+        // We want those include paths to come from our `zig cc` instead of the local clang.
+        // There are three main mechanisms possible:
+        //
+        // 1. Replace the local clang with our version.
+        //
+        //    Bindgen, internally via clang-sys, recognizes `CLANG_PATH` and `PATH`.
+        //    They are unfortunately a global namespace and simply setting them may break
+        //    existing build scripts, so we can't confidently override them.
+        //
+        //    Clang-sys can also look at target-prefixed clang if arguments contain `-target`.
+        //    Unfortunately clang-sys can only recognize `-target xxx`, which very slightly
+        //    differs from what bindgen would pass (`-target=xxx`), so this is not yet possible.
+        //
+        //    It should be also noted that we need to collect not only include paths
+        //    but macro definitions added by Zig, for example `-D__GLIBC_MINOR__`.
+        //    Clang-sys can't do this yet, so this option seems less robust than we want.
+        //
+        // 2. Set the environment variable `BINDGEN_EXTRA_CLANG_ARGS` and let bindgen to
+        //    append them to arguments passed to `libclang.so`.
+        //
+        //    This unfortunately means that we have the same set of arguments for C and C++.
+        //    Also we have to support older versions of clang, as old as clang 5 (2017).
+        //    We do have options like `-c-isystem` (cc1 only) and `-cxx-isystem`,
+        //    but we need to be aware of other options may affect our added options
+        //    and this requires a nitty gritty of clang driver and cc1---really annoying.
+        //
+        // 3. Fix either bindgen or clang-sys or Zig to ease our jobs.
+        //
+        //    This is not the option for now because, even after fixes, we have to support
+        //    older versions of bindgen or Zig which won't have those fixes anyway.
+        //    But it seems that minor changes to bindgen can indeed fix lots of issues
+        //    we face, so we are looking for them in the future.
+        //
+        // For this reason, we chose the option 2 and overrode `BINDGEN_EXTRA_CLANG_ARGS`.
+        // The following therefore assumes some understanding about clang option handling,
+        // including what the heck is cc1 (see the clang FAQ) and how driver options get
+        // translated to cc1 options (no documentation at all, as it's supposedly unstable).
+        // Fortunately for us, most (but not all) `-i...` options are passed through cc1.
+        //
+        // If you do experience weird compilation errors during bindgen, there's a chance
+        // that this code has overlooked some edge cases. You can put `.clang_arg("-###")`
+        // to print the final cc1 options, which would give a lot of information about
+        // how it got screwed up and help a lot when we fix the issue.
+        //
+        // </digression>
+
+        let mut args = Vec::new();
+
+        // Never include default include directories,
+        // otherwise `__has_include` will be totally confused.
+        args.push("-nostdinc".to_owned());
+
+        // Add various options for libc++ and glibc.
+        // Should match what `Compilation.zig` internally does:
+        //
+        // https://github.com/ziglang/zig/blob/0.9.0/src/Compilation.zig#L3390-L3427
+        // https://github.com/ziglang/zig/blob/0.9.1/src/Compilation.zig#L3408-L3445
+        // https://github.com/ziglang/zig/blob/0.10.0/src/Compilation.zig#L4163-L4211
+        // https://github.com/ziglang/zig/blob/0.10.1/src/Compilation.zig#L4240-L4288
+        if raw_target.contains("musl") {
+            args.push("-D_LIBCPP_HAS_MUSL_LIBC".to_owned());
+        }
+        args.push("-D_LIBCPP_DISABLE_VISIBILITY_ANNOTATIONS".to_owned());
+        args.push("-D_LIBCXXABI_DISABLE_VISIBILITY_ANNOTATIONS".to_owned());
+        args.push("-D_LIBCPP_HAS_NO_VENDOR_AVAILABILITY_ANNOTATIONS".to_owned());
+        args.push("-D_LIBCPP_ABI_VERSION=1".to_owned());
+        args.push("-D_LIBCPP_ABI_NAMESPACE=__1".to_owned());
+        if let Some(ver) = c_opts.glibc_minor_ver {
+            // Handled separately because we have no way to infer this without Zig
+            args.push(format!("-D__GLIBC_MINOR__={ver}"));
+        }
+
+        for (kind, path) in cpp_paths.drain(..cpp_pre_len) {
+            assert!(kind == Kind::Normal);
+            // Ideally this should be `-stdlib++-isystem`, which can be disabled by
+            // passing `-nostdinc++`, but it is fairly new: https://reviews.llvm.org/D64089
+            //
+            // (Also note that `-stdlib++-isystem` is a driver-only option,
+            // so it will be moved relative to other `-isystem` options against our will.)
+            args.push("-cxx-isystem".to_owned());
+            args.push(path);
+        }
+
+        for (kind, path) in c_paths {
+            match kind {
+                Kind::Normal => {
+                    // A normal `-isystem` is preferred over `-cxx-isystem` by cc1...
+                    args.push("-Xclang".to_owned());
+                    args.push("-c-isystem".to_owned());
+                    args.push("-Xclang".to_owned());
+                    args.push(path.clone());
+                    args.push("-cxx-isystem".to_owned());
+                    args.push(path);
+                }
+                Kind::Framework => {
+                    args.push("-iframework".to_owned());
+                    args.push(path);
+                }
+            }
+        }
+
+        for (kind, path) in cpp_paths.drain(cpp_paths.len() - cpp_post_len..) {
+            assert!(kind == Kind::Normal);
+            args.push("-cxx-isystem".to_owned());
+            args.push(path);
+        }
+
+        Ok(args)
+    }
+
     fn setup_os_deps(
         manifest_path: Option<&Path>,
         release: bool,
@@ -505,9 +775,10 @@ impl Zig {
                         // cargo install doesn't pass a manifest path so `Cargo.toml` in cwd may not exist
                         continue;
                     }
-                    let mut metadata_cmd = cargo_metadata::MetadataCommand::new();
-                    metadata_cmd.manifest_path(manifest_path);
-                    let metadata = metadata_cmd.exec()?;
+                    let metadata = cargo_metadata::MetadataCommand::new()
+                        .manifest_path(manifest_path)
+                        .no_deps()
+                        .exec()?;
                     metadata.target_directory.into_std_path_buf().join(target)
                 };
                 let profile = match cargo.profile.as_deref() {
@@ -524,7 +795,19 @@ impl Zig {
                 };
                 let deps_dir = target_dir.join(profile).join("deps");
                 fs::create_dir_all(&deps_dir)?;
+                if !target_dir.join("CACHEDIR.TAG").is_file() {
+                    // Create a CACHEDIR.TAG file to exclude target directory from backup
+                    let _ = fs::write(
+                        target_dir.join("CACHEDIR.TAG"),
+                        "Signature: 8a477f597d28d172789f06886806bc55
+# This file is a cache directory tag created by cargo.
+# For information about cache directory tags see https://bford.info/cachedir/
+",
+                    );
+                }
                 fs::write(deps_dir.join("libiconv.tbd"), LIBICONV_TBD)?;
+                fs::write(deps_dir.join("libcharset.1.tbd"), LIBCHARSET_TBD)?;
+                fs::write(deps_dir.join("libcharset.tbd"), LIBCHARSET_TBD)?;
             } else if target.contains("arm") && target.contains("linux") {
                 // See https://github.com/ziglang/zig/issues/3287
                 if let Ok(lib_dir) = Zig::lib_dir() {
@@ -545,7 +828,8 @@ impl Zig {
                     if !synchronization_def.is_file() {
                         let api_ms_win_core_synch_l1_2_0_def =
                             lib_common.join("api-ms-win-core-synch-l1-2-0.def");
-                        fs::copy(api_ms_win_core_synch_l1_2_0_def, synchronization_def)?;
+                        // Ignore error
+                        fs::copy(api_ms_win_core_synch_l1_2_0_def, synchronization_def).ok();
                     }
                 }
             }
@@ -591,12 +875,15 @@ set(CMAKE_CXX_COMPILER {cxx})
 set(CMAKE_RANLIB {ranlib})"#,
             system_name = system_name,
             system_processor = system_processor,
-            cc = zig_wrapper.cc.display(),
-            cxx = zig_wrapper.cxx.display(),
-            ranlib = zig_wrapper.ranlib.display(),
+            cc = zig_wrapper.cc.to_slash_lossy(),
+            cxx = zig_wrapper.cxx.to_slash_lossy(),
+            ranlib = zig_wrapper.ranlib.to_slash_lossy(),
         );
         if enable_zig_ar {
-            content.push_str(&format!("\nset(CMAKE_AR {})\n", zig_wrapper.ar.display()));
+            content.push_str(&format!(
+                "\nset(CMAKE_AR {})\n",
+                zig_wrapper.ar.to_slash_lossy()
+            ));
         }
         fs::write(&toolchain_file, content)?;
         Ok(toolchain_file)
@@ -699,6 +986,7 @@ pub fn prepare_zig_linker(target: &str) -> Result<ZigWrapper> {
     let arch = triple.architecture.to_string();
     let target_env = match (triple.architecture, triple.environment) {
         (Architecture::Mips32(..), Environment::Gnu) => Environment::Gnueabihf,
+        (Architecture::Powerpc, Environment::Gnu) => Environment::Gnueabihf,
         (_, Environment::GnuLlvm) => Environment::Gnu,
         (_, environment) => environment,
     };
@@ -766,14 +1054,25 @@ pub fn prepare_zig_linker(target: &str) -> Result<ZigWrapper> {
             };
             format!("-target {zig_arch}-windows-{target_env}{abi_suffix} {cc_args}",)
         }
+        OperatingSystem::Emscripten => format!("-target {arch}-emscripten{abi_suffix} {cc_args}"),
+        OperatingSystem::Wasi => format!("-target {arch}-wasi{abi_suffix} {cc_args}"),
+        OperatingSystem::Unknown => {
+            if triple.architecture == Architecture::Wasm32
+                || triple.architecture == Architecture::Wasm64
+            {
+                format!("-target {arch}-freestanding{abi_suffix} {cc_args}")
+            } else {
+                bail!("unsupported target '{rust_target}'")
+            }
+        }
         _ => bail!(format!("unsupported target '{rust_target}'")),
     };
 
     let zig_linker_dir = cache_dir();
     fs::create_dir_all(&zig_linker_dir)?;
 
-    if triple.operating_system == OperatingSystem::Linux
-        && matches!(
+    if triple.operating_system == OperatingSystem::Linux {
+        if matches!(
             triple.environment,
             Environment::Gnu
                 | Environment::Gnuspe
@@ -782,35 +1081,60 @@ pub fn prepare_zig_linker(target: &str) -> Result<ZigWrapper> {
                 | Environment::Gnuabi64
                 | Environment::GnuIlp32
                 | Environment::Gnueabihf
-        )
-    {
-        let glibc_version = if abi_suffix.is_empty() {
-            (2, 17)
-        } else {
-            let mut parts = abi_suffix[1..].split('.');
-            let major: usize = parts.next().unwrap().parse()?;
-            let minor: usize = parts.next().unwrap().parse()?;
-            (major, minor)
-        };
-        // See https://github.com/ziglang/zig/issues/9485
-        if glibc_version < (2, 28) {
-            use crate::linux::{FCNTL_H, FCNTL_MAP};
+        ) {
+            let glibc_version = if abi_suffix.is_empty() {
+                (2, 17)
+            } else {
+                let mut parts = abi_suffix[1..].split('.');
+                let major: usize = parts.next().unwrap().parse()?;
+                let minor: usize = parts.next().unwrap().parse()?;
+                (major, minor)
+            };
+            // See https://github.com/ziglang/zig/issues/9485
+            if glibc_version < (2, 28) {
+                use crate::linux::{FCNTL_H, FCNTL_MAP};
+                use std::fmt::Write as _;
+
+                let zig_version = Zig::zig_version()?;
+                if zig_version.major == 0 && zig_version.minor < 11 {
+                    let fcntl_map = zig_linker_dir.join("fcntl.map");
+                    fs::write(&fcntl_map, FCNTL_MAP)?;
+                    let fcntl_h = zig_linker_dir.join("fcntl.h");
+                    fs::write(&fcntl_h, FCNTL_H)?;
+
+                    write!(
+                        cc_args,
+                        " -Wl,--version-script={} -include {}",
+                        fcntl_map.display(),
+                        fcntl_h.display()
+                    )
+                    .unwrap();
+                }
+            }
+        } else if matches!(
+            triple.environment,
+            Environment::Musl
+                | Environment::Muslabi64
+                | Environment::Musleabi
+                | Environment::Musleabihf
+        ) {
+            use crate::linux::MUSL_WEAK_SYMBOLS_MAPPING_SCRIPT;
             use std::fmt::Write as _;
 
             let zig_version = Zig::zig_version()?;
-            if zig_version.major == 0 && zig_version.minor < 11 {
-                let fcntl_map = zig_linker_dir.join("fcntl.map");
-                fs::write(&fcntl_map, FCNTL_MAP)?;
-                let fcntl_h = zig_linker_dir.join("fcntl.h");
-                fs::write(&fcntl_h, FCNTL_H)?;
+            let rustc_version = rustc_version::version_meta()?.semver;
 
-                write!(
-                    cc_args,
-                    " -Wl,--version-script={} -include {}",
-                    fcntl_map.display(),
-                    fcntl_h.display()
-                )
-                .unwrap();
+            // as zig 0.11.0 is released, its musl has been upgraded to 1.2.4 with break changes
+            // but rust is still with musl 1.2.3
+            // we need this workaround before rust 1.72
+            // https://github.com/ziglang/zig/pull/16098
+            if (zig_version.major, zig_version.minor) >= (0, 11)
+                && (rustc_version.major, rustc_version.minor) < (1, 72)
+            {
+                let weak_symbols_map = zig_linker_dir.join("musl_weak_symbols_map.ld");
+                fs::write(&weak_symbols_map, MUSL_WEAK_SYMBOLS_MAPPING_SCRIPT)?;
+
+                write!(cc_args, " -Wl,-T,{}", weak_symbols_map.display()).unwrap()
             }
         }
     }
